# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

"""The ingress point of the application."""

import optparse
import os
import stat
import sys
import glib
import gtk
import math
from a4lib.presentation import PresentationError, NoMetadataError
from a4lib.player import GtkCairoPlayer
from a4lib.editor import Editor

UI_PATHS = ('data', '/usr/local/share/a4/data', '/usr/share/a4/data')


class WindowMain(object):
    """The main window of the application."""

    def __init__(self):
        self.builder = gtk.Builder()
        for path in UI_PATHS:
            file_name = os.path.join(path, 'window_main.glade')
            try:
                self.builder.add_from_file(file_name)
            except glib.GError as error:
                pass
            else:
                break
        else:
            sys.exit('error: cannot open UI file: {0}'.format(error.args[0]))

        self.builder.connect_signals(self)
        area = self.builder.get_object('drawing_area')
        area.connect('key-press-event', self.on_area_keypress)
        ## This is for telling the drawing area to notify mouse events
        ## and must be done *before* calling show() method.
        area.set_events(gtk.gdk.EXPOSURE_MASK
                            | gtk.gdk.LEAVE_NOTIFY_MASK
                            | gtk.gdk.BUTTON_PRESS_MASK
                            | gtk.gdk.BUTTON_RELEASE_MASK
                            | gtk.gdk.POINTER_MOTION_MASK
                            | gtk.gdk.POINTER_MOTION_HINT_MASK)
        self.gtk_window = self.builder.get_object('window_main')
        self.gtk_window.show()
        self.drawing_area = area
        self.reset()

        self.player = None
        self.last_path = None
        self.is_fullscreen = False
        self.callback_called = False

    def set_status(self, status):
        """Put the given string in the status bar."""
        label = self.builder.get_object('label_status')
        label.set_text(status)

    def open_file(self, file_name, is_edited=False, force_loading=False):
        """Open a file inside this window."""
        self.set_status('Loading {0}...'.format(file_name))
        try:
            # Try to open a file.
            if is_edited:
                self.player = Editor(file_name, self.drawing_area,
                        self.builder.get_object('iconview1'), force_loading)
            else:
                self.player = GtkCairoPlayer(file_name, self.drawing_area,
                        force_loading)
        except PresentationError as error:
            # The file doesn't exist, or is not an SVG file. Show an error
            # dialog and don't do anything else.
            dialog = gtk.MessageDialog(
                self.gtk_window, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE,
                    'Cannot open {0!r}: {1[0]}'.format(file_name, error.args))
            dialog.run()
            dialog.destroy()
        except NoMetadataError as error:
            msg = 'No A4 data found into file,\nwould you like to convert this file into an A4 presentation?'
            dialog = gtk.Dialog('Import', self.gtk_window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                         gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
            dialog.vbox.pack_start(gtk.Label(msg))
            dialog.show_all()
            result = dialog.run()
            if result == gtk.RESPONSE_ACCEPT:
                self.open_file(file_name, is_edited, True)
            dialog.destroy()
        else:
            # The file was OK. Set up the environment.
            # Make the buttons sensitive.
            for button in ('imagemenuitem_close', 'toolbutton_close',
                           'toolbutton_next', 'toolbutton_zoom_in',
                           'toolbutton_zoom_out', 'toolbutton_editor',
                           'spin_rotation', 'imagemenuitem_save',
                           'imagemenuitem_saveas'):
                self.builder.get_object(button).set_sensitive(True)
            self.builder.get_object('scrolledwindow1').set_visible(is_edited)
            spin_button = self.builder.get_object('spin_rotation')
            spin_button.set_value(0)
            ## checking file writability
            mode = stat.S_IMODE(os.stat(file_name).st_mode)
            writable = mode & (stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
            if writable:
                self.builder.get_object('toolbutton_save').set_sensitive(True)

            def change_rotation(region):
                degrees = int(math.degrees(region.rotate))
                self.callback_called = True
                spin_button.set_value(math.degrees(region.rotate))
            self.player.set_region_changed_callback(change_rotation)
        finally:
            # Restore the status string.
            self.set_status('')

    def reset(self):
        """Reset the environment and put the window into its original state.

        This function closes the file (if opened) and resets all the buttons
        and the menus.
        """
        # Destroy the image object and reset the drawing area.
        self.player = None
        ## FIXME: it'd be good if CairoPlayer will take care of it:
        self.drawing_area.queue_draw()
        for button in ('imagemenuitem_close', 'toolbutton_close',
                       'toolbutton_previous', 'toolbutton_next',
                       'toolbutton_zoom_in', 'toolbutton_zoom_out',
                       'toolbutton_editor', 'spin_rotation',
                       'toolbutton_save', 'imagemenuitem_save',
                       'imagemenuitem_saveas'):
            self.builder.get_object(button).set_sensitive(False)
        self.builder.get_object('spin_rotation').set_value(0)
        self.builder.get_object('toolbutton_editor').set_active(0)

    def on_open_clicked(self, widget):
        """Event called when the 'Open' button is clicked."""
        # Set up the file chooser dialog.
        dialog = gtk.FileChooserDialog(
            None, None, gtk.FILE_CHOOSER_ACTION_OPEN, (
                gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        # If an another file has been opened, use its location as starting
        # folder for the dialog.
        if self.last_path is not None:
            dialog.set_current_folder(path)

        # Show the dialog, back up the relevant properties and destroy it.
        response = dialog.run()
        file_name = dialog.get_filename()
        dialog.destroy()
        # Open the file, if necessary.
        if response == gtk.RESPONSE_OK:
            self.reset()
            self.open_file(file_name)

    def on_save_clicked(self, widget):
        if self.player:
            self.player.presentation._save_path()

    def on_save_as_clicked(self, widget):
        dialog = gtk.FileChooserDialog(
            None, None, gtk.FILE_CHOOSER_ACTION_SAVE, (
                gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        # If an another file has been opened, use its location as starting
        # folder for the dialog.
        if self.last_path is not None:
            dialog.set_current_folder(path)

        # Show the dialog, back up the relevant properties and destroy it.
        response = dialog.run()
        file_name = dialog.get_filename()
        dialog.destroy()
        # Save in the specified location
        if response == gtk.RESPONSE_OK and self.player:
            self.player.presentation._save_path(file_name)

    def on_close_clicked(self, widget):
        """The 'Close' button has been clicked."""
        self.reset()

    def on_next_clicked(self, widget):
        """The 'Next' button has been clicked."""
        if self.player:
            self.player.next()
            self.on_roi_changed()

    def on_previous_clicked(self, widget):
        """The 'Previous' button has been clicked."""
        if self.player:
            self.player.previous()
            self.on_roi_changed()

    def on_fullscreen_activate(self, widget):
        """ Toggle fullscreen status """
        if not self.is_fullscreen:
            self.gtk_window.fullscreen()
            self.is_fullscreen = True
        else:
            self.gtk_window.unfullscreen()
            self.is_fullscreen = False
        #self.builder.get_object('menubar1').set_visible(not self.is_fullscreen)
        #self.builder.get_object('main_toolbar').set_visible(not self.is_fullscreen)

    def on_zoom_in_clicked(self, widget):
        "The zoom in button has been clicked"
        if self.player != None:
            self.player.zoom_in()

    def on_zoom_out_clicked(self, widget):
        "The zoom out button has been cliched"
        if self.player != None:
            self.player.zoom_out()

    def on_editor_clicked(self, widget):
        "Editor/Player switch has been clicked"
        if self.player != None:
            ## saving the current region and view_status to keep it between
            ## mode switches.
            region = self.player.current_region
            view_status = self.player.view_status
            ## switch between Player and Editor class
            self.open_file(self.player.presentation.file_name,
                not isinstance(self.player, Editor))
            self.player.current_region = region
            self.player.view_status = view_status
            self.player.current_animation = None
            self.drawing_area.queue_draw()
            if isinstance(self.player, Editor):
                widget.set_active(True)
                self.builder.get_object('drawing_toolbar').set_visible(True)
            else:
                widget.set_active(False)
                self.builder.get_object('drawing_toolbar').set_visible(False)

    def _ask_image_dialog(self):
        dialog = gtk.FileChooserDialog(
            None, None, gtk.FILE_CHOOSER_ACTION_OPEN, (
                gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        # If an another file has been opened, use its location as starting
        # folder for the dialog.
        if self.last_path is not None:
            dialog.set_current_folder(path)

        # Show the dialog, back up the relevant properties and destroy it.
        response = dialog.run()
        file_name = dialog.get_filename()
        dialog.destroy()
        # Save in the specified location
        if response == gtk.RESPONSE_OK and self.player:
            return file_name
        return ""

    def on_toggle_tool(self, widget):
        if not isinstance(self.player, Editor):
            return
        toggle_modes = {'SelectToolbutton': 'select',
                        'PathToolbutton': 'path',
                        'ShapesToolbutton': 'rect',
                        'ImageToolbutton': 'image',
                        'PolaroidToolbutton': 'polaroid',
                        'TextToolbutton': 'text'}

        widget_name = gtk.Buildable.get_name(widget)
        if widget.get_active():
            for i in toggle_modes.keys():
                if i != widget_name:
                    obj = self.builder.get_object(i)
                    if obj.get_active():
                        obj.set_active(False)
            object_name = toggle_modes[widget_name]
            if object_name in ['path', 'select']:
                self.player.set_editor_mode(object_name)
            else:
                self.player.set_editor_mode('objects')
            self.player.set_next_new_object(toggle_modes[widget_name])
            if toggle_modes[widget_name] == 'image':
                self.player.set_target_image(self._ask_image_dialog())
        if not any(self.builder.get_object(i).get_active() for i in toggle_modes.keys()):
            self.player.set_editor_mode()
        path_widget = self.builder.get_object('PathToolbutton')
        self.player.set_show_path(path_widget.get_active())
        self.drawing_area.queue_draw()

    def on_drawing_area_expose(self, widget, event):
        """This method is called everytime the drawing area should be redrawn.
        """
        if self.player is None:
            return
        context = self.drawing_area.window.cairo_create()

        context.rectangle(
            event.area.x, event.area.y, event.area.width, event.area.height)
        context.clip()
        context.set_source_rgb(1, 1, 1)
        context.paint()

        self.player.render(context)

    def on_area_keypress(self, widget, event):
        """Callback for the key press event."""
        key = gtk.gdk.keyval_name(event.keyval)
        if isinstance(self.player, Editor):
            if self.player.on_area_key_press(widget, event):
                return
        if key == 'Left':
            self.on_previous_clicked(None)
        elif key == 'Right':
            self.on_next_clicked(None)
        elif key in '0123456789':
            # XXX I think most of the presentation will have more than 10
            # regions. Probably it would be better to give the user the ability
            # to decide which keys to associate to a given presentation.
            # [bug=605310]
            self.player.go_to(int(key))
            self.on_roi_changed()
        elif key == 'minus':
            self.player.zoom_out()
        elif key == 'plus':
            self.player.zoom_in()

    def on_roi_changed(self):
        """Set buttons sensitivity according to the current ROI"""
        button = self.builder.get_object('toolbutton_previous')
        button.set_sensitive(not self.player.first_region)
        button = self.builder.get_object('toolbutton_next')
        button.set_sensitive(not self.player.last_region)

    def on_about_menu(self, widget):
        """Show the About menu"""
        from a4lib import __version__
        dialog = gtk.AboutDialog()
        dialog.set_name('A4')
        dialog.set_version(__version__)
        dialog.set_comments('A stylish tool to create and view cute presentations.')
        dialog.set_website('http://launchpad.net/a4')
        dialog.set_website_label('A4 website')
        dialog.set_copyright('Copyright (C) 2010 A4 Developers')
        #dialog.set_logo("LOGO_GOES_HERE")
        #dialog.set_icon_from_file("ICON_GOES_HERE")
        #code to automagically fetch authors and license should be used here
        dialog.set_authors(['A4 Developers'])
        dialog.set_license('A4 is released under the terms of GNU GPL, version 3.')
        dialog.run()
        dialog.destroy()

    def on_spin_rotation_value_changed(self, widget):
        if not self.player:
            return None
        if self.callback_called:
            self.callback_called = False
            return
        radians = math.radians(widget.get_value())
        self.player.rotate(radians)

    def quit(self, widget=None):
        """Close the window and quit the application."""
        gtk.main_quit()


def parse_arguments(args=None):
    """Parse the command line arguments."""
    parser = optparse.OptionParser()
    return parser.parse_args(args)


def main(args=None):
    """Start the application."""
    options, files = parse_arguments(args)
    window = WindowMain()
    if files:
        # TODO What if more than one file is given?
        window.open_file(files[0])
    gtk.main()
