# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

import cairo
import gtk
import math
from numpy import array, dot
from numpy.linalg import inv
from a4lib.player import GtkCairoPlayer
from a4lib.region import Region

from presentation_objects import RectObject, TextObject, PolaroidObject, ImageObject


def object_class_from_name(name):
    obj_dict = {'rect': RectObject, 'text': TextObject,
                'image': ImageObject, 'polaroid': PolaroidObject}
    return obj_dict[name]


def cairo_draw_rect(ctx, x0, y0, x1, y1, r, g, b, alpha):
        ctx.set_source_rgb(0, 0xff, 0)
        ctx.move_to(x0, y0)
        ctx.line_to(x1, y0)
        ctx.line_to(x1, y1)
        ctx.line_to(x0, y1)
        ctx.line_to(x0, y0)
        ctx.set_line_width(0.2)
        ctx.stroke()
        ctx.rectangle(x0, y0, x1 - x0, y1 - y0)
        ctx.set_source_rgba(0, 0, 0.5, 0.50)
        ctx.fill()


class PresentationPathHandler(object):
    def __init__(self, presentation, treeview):
        self.treeview = treeview
        self.presentation = presentation
        self.model = gtk.ListStore(int, str, gtk.gdk.Pixbuf)

        self.treeview.set_model(self.model)
        self.treeview.set_text_column(1)
        self.treeview.set_pixbuf_column(2)

        self.treeview.set_columns(1)
        self.treeview.set_item_width(120)
        self.treeview.set_size_request(130, -1)

    def draw_region_on_iconview(self, region, position=0):
        pixbuf_height = abs(int(100 * region.height / region.width))
        pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, 100,
                    pixbuf_height)
        pix_data = pixbuf.get_pixels_array()
        pixbuf_size = (pixbuf.get_width(), pixbuf.get_height())
        surface = cairo.ImageSurface.create_for_data(pix_data, cairo.FORMAT_ARGB32,
                     pixbuf_size[0], pixbuf_size[1], pixbuf.get_rowstride())
        cr = cairo.Context(surface)
        region.render(cr, self.presentation, pixbuf_size)

        self.model.append([position, str(position), pixbuf])

    def draw_regions_on_iconview(self, regions):
        for n, region in enumerate(regions):
            self.draw_region_on_iconview(region, n)

    def highlight_path_region(self, context, region):
        context.save()
        context.translate(region.xc, region.yc)
        context.rotate(region.rotate)
        context.translate(-region.xc, -region.yc)
        x0 = region.xc - region.width / 2
        y0 = region.yc - region.height / 2
        x1 = region.xc + region.width / 2
        y1 = region.yc + region.height / 2
        cairo_draw_rect(context, x0, y0, x1, y1, 0, 0, 0.5, 0.50)
        context.restore()

    def render(self, context):
        for i in xrange(0, len(self.presentation.path)):
            region = self.presentation.get_path_region(i)
            self.highlight_path_region(context, region)


class Editor(GtkCairoPlayer):

    def __init__(self, file_name, drawing_area, treeview, force_loading=False):
        GtkCairoPlayer.__init__(self, file_name, drawing_area, force_loading)
        self.path_handler = PresentationPathHandler(self.presentation, treeview)
        ## used to keep status of mouse actions:
        ## value currently used: [None,'start','moving']
        self.action_status = None
        ## tells if we should draw the rect of the path.
        self.show_path = None
        ## tell the tool we're using and how we should behave.
        self.editor_modes = ['select', 'path', 'objects']
        self.editor_mode = 'select'
        self.new_object_name = None
        ## the selected object id:
        self.current_object = None
        self.target_filename = None

        ph = self.path_handler
        ph.draw_regions_on_iconview(self.presentation.get_path_regions())

    def set_show_path(self, flag=True):
        self.show_path = flag

    # objects mode: rect/polaroid
    def set_next_new_object(self, obj_name):
        self.new_object_name = obj_name

    # select/objects
    def set_editor_mode(self, status='select'):
        if status in self.editor_modes:
            self.editor_mode = status

    def set_target_image(self, file_name):
        self.target_filename = file_name

    def on_area_button_press_event(self, widget, event):
        event_point = (event.x, event.y)
        winsize = self.area.window.get_size()
        cr = self.current_region
        point = cr.get_point_from_window_coordinates(winsize, event_point)

        self.selection_start_position = (event.x, event.y)
        self.action_status = 'start'
        if self.editor_mode == 'select':
            pass  # presentation.select(something)
        elif self.editor_mode == 'objects':
            cls = object_class_from_name(self.new_object_name)
            self.current_object = cls(x=point[0], y=point[1],
                                      rotate=self.current_region.rotate)
            if self.new_object_name == 'image':
                self.current_object.set_image(self.target_filename)
            self.presentation.add_object(self.current_object)
        else:
            GtkCairoPlayer.on_area_button_press_event(self, widget, event)

    def on_area_button_release_event(self, widget, event):
        GtkCairoPlayer.on_area_button_release_event(self, widget, event)
        if self.editor_mode == 'path' and event.state & gtk.gdk.CONTROL_MASK:
            rect = (self.selection_start_position, self.mouse_position)
            self.add_treeview_icon(rect)
        elif self.editor_mode == 'objects':
            pass
        self.action_status = None

    def on_area_motion_notify_event(self, widget, event):
        if self.action_status is None:
            GtkCairoPlayer.on_area_motion_notify_event(self, widget, event)
        else:
            if self.editor_mode == 'path':
                if event.state & gtk.gdk.CONTROL_MASK:
                    self.mouse_position = (event.x, event.y)
                    self.area.queue_draw()
            elif self.editor_mode == 'objects' and self.action_status:
                if self.current_object:
                    winsize = self.area.window.get_size()
                    cr = self.current_region
                    scale = self.current_region.get_window_scale_factor(winsize)
                    w = (event.x - self.selection_start_position[0]) / scale
                    h = (event.y - self.selection_start_position[1]) / scale
                    self.current_object.set_size(w, h)
                self.mouse_position = (event.x, event.y)
                self.area.queue_draw()
            self.action_status = 'moving'

    def on_area_key_press(self, widget, event):
        if self.current_object:
            if True:  # premo tutto tranne esc:
                key = gtk.gdk.keyval_name(event.keyval)
                self.current_object.character_pressed(key)
                self.area.queue_draw()
            elif False:  # premo esc
                self.action_status = None
            return True

    def draw_selection_box(self, context):
        win_size = self.area.window.get_size()
        cr = self.current_region
        x0 = self.selection_start_position[0]
        y0 = self.selection_start_position[1]
        x1 = self.mouse_position[0]
        y1 = self.mouse_position[1]
        x0, y0 = cr.get_point_from_window_coordinates(win_size, (x0, y0))
        x1, y1 = cr.get_point_from_window_coordinates(win_size, (x1, y1))
        cairo_draw_rect(context, x0, y0, x1, y1, 0, 0, 0.5, 0.50)

    def render(self, context):
        #   render a transparent box (should indicate the selection)
        # render already existant path points.
        # render an icon for rotating the whole image.
        GtkCairoPlayer.render(self, context)
        if self.show_path:
            ## render the path over the presentation
            self.path_handler.render(context)
            ## render the selection (TODO: not only for path)
            if self.editor_mode == 'path' and self.action_status:
                self.draw_selection_box(context)

    def add_treeview_icon(self, rect=None):
        ## convert a rect into a Region()
        win_size = self.area.window.get_size()

        cr = self.current_region
        rect2 = (cr.get_point_from_window_coordinates(win_size, rect[0]),
                cr.get_point_from_window_coordinates(win_size, rect[1]))
        xc = (rect2[0][0] + rect2[1][0]) / 2
        yc = (rect2[0][1] + rect2[1][1]) / 2
        region = Region(xc, yc,
                    (rect2[1][0] - rect2[0][0]),
                    (rect2[1][1] - rect2[0][1]),
                    cr.rotate)
        ## * metti la regione in lista
        self.presentation.path.append(region)
        ## * usa la regione per la pixbuf.
        self.path_handler.draw_region_on_iconview(region, len(self.presentation.path) - 1)
