 # Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

import glib
import gtk
from a4lib.presentation import Presentation, FeaturedPresentation
from a4lib.region import Region, Animation, linear_interpolation


class Player(object):
    """Handle the frame moving generically from a position to another."""

    def __init__(self, file_name, force_loading=False):
        image = FeaturedPresentation(file_name, force_loading)
        self.presentation = image
        self._current_region = Region.whole_image(image)
        self.current_animation = None
        # Each zoom call Region.scale_of with this factor.
        self.zoom_factor = 1.4
        self.view_status = -1

        self.region_changed_callback = None

    def get_current_region(self):
        return self._current_region

    def set_current_region(self, region):
        self._current_region = region
        if self.region_changed_callback:
            self.region_changed_callback(region)

    current_region = property(get_current_region, set_current_region)

    def set_region_changed_callback(self, callback):
        self.region_changed_callback = callback

    def _set_animation_to(self, target):
        """Calculate `Animation` given the id of two point of a presentation.
        """
        if target < 0:
            target_region = Region.whole_image(self.presentation)
        else:
            image = self.presentation
            target_region = image.get_path_region(target)
        self.current_animation = Animation(self.current_region, target_region)

    def next(self):
        """Send the frame to the next presentation region."""
        self.view_status = min(
            self.view_status + 1, len(self.presentation.path) - 1)
        self._set_animation_to(self.view_status)

    def previous(self):
        """Send the frame to the previous presentation region."""
        if self.view_status > -1:
            self.view_status -= 1
        self._set_animation_to(self.view_status)

    def go_to(self, position):
        self.view_status = min(
            position, len(self.presentation.path) - 1)
        self._set_animation_to(self.view_status)

    def zoom_in(self, translation=(0, 0)):
        if self.current_region is None:
            return
        factor = 1.0 / self.zoom_factor
        region = self.current_region.scale_of(
            factor, factor).move(*translation)
        self.current_animation = Animation(
            self.current_region, region, 0.2, linear_interpolation)

    def zoom_out(self, translation=(0, 0)):
        if self.current_region is None:
            return
        region = self.current_region.scale_of(
            self.zoom_factor, self.zoom_factor).move(*translation)
        self.current_animation = Animation(
            self.current_region, region, 0.2, linear_interpolation)

    def move(self, step_x, step_y):
        if self.current_animation and self.current_animation.status != 'ended':
            self.current_animation = None
        self.current_region = self.current_region.move(step_x, step_y)

    def rotate(self, rotate):
        region = Region(self.current_region.xc, self.current_region.yc,
                        self.current_region.width, self.current_region.height,
                        rotate)
        self.current_animation = Animation(
            self.current_region, region, 0.2, linear_interpolation)

    @property
    def last_region(self):
        return self.view_status == len(self.presentation.path) - 1

    @property
    def first_region(self):
        return self.view_status == -1


class CairoPlayer(Player):
    """This class extends `Player` with the cairo rendering functions."""

    def __init__(self, file_name, drawing_area, force_loading=False):
        Player.__init__(self, file_name, force_loading)
        self.area = drawing_area
        drawing_area.queue_draw()

    def _set_timer(self, interval=100):
        self.area.queue_draw()
        self.timer = glib.timeout_add(interval, self.tick)

    def next(self):
        Player.next(self)
        self._set_timer()

    def previous(self):
        Player.previous(self)
        self._set_timer()

    def tick(self):
        if self.current_animation and self.current_animation.status != 'ended':
            self.area.queue_draw()
            return True
        self.current_animation = None
        glib.source_remove(self.timer)

    def go_to(self, position):
        Player.go_to(self, position)
        self._set_timer()

    def zoom_in(self, translation=(0, 0)):
        Player.zoom_in(self, translation)
        self._set_timer()

    def zoom_out(self, translation=(0, 0)):
        Player.zoom_out(self, translation)
        self._set_timer()

    def rotate(self, radians):
        Player.rotate(self, radians)
        self._set_timer()

    def move(self, stepx, stepy):
        win_size = self.area.window.get_size()
        scale = self.current_region.get_window_scale_factor(win_size)
        translate = self.current_region.get_window_fit_vector(win_size)

        stepx = (self.current_region.width + translate[0] / scale) * (
            stepx / win_size[0])
        stepy = (self.current_region.height + translate[1] / scale) * (
            stepy / win_size[1])
        Player.move(self, stepx, stepy)
        self.area.queue_draw()

    def render(self, context):
        """Render a frame, considering the current point of an animation."""
        if not self.presentation:
            return
        if self.current_animation and self.current_animation.status != 'ended':
            self.current_region = self.current_animation.get_next_region()

        self.current_region.render(
            context, self.presentation, self.area.window.get_size())


class GtkCairoPlayer(CairoPlayer):
    """This class extends `CairoPlayer` with the handling of GTK events."""

    def __init__(self, file_name, area, force_loading=False):
        CairoPlayer.__init__(self, file_name, area, force_loading)
        area.connect('button_press_event', self.on_area_button_press_event)
        area.connect('button_release_event', self.on_area_button_release_event)
        area.connect('scroll_event', self.on_area_scroll_event)
        area.connect('motion-notify-event', self.on_area_motion_notify_event)

        self.mouse_pressed = False
        self.mouse_position = (0, 0)

    def _mouse_step_factor(self, factor):
        # Resize factor, between GTK window and the SVG images.
        current_region = self.current_region
        resize_factor = current_region.width / self.area.window.get_size()[0]

        # Pointer position adaptation.
        mouse_step = (
            self.mouse_position[0] * resize_factor,
            self.mouse_position[1] * resize_factor)
        mouse_step_out = [
            mouse_step[0] - (mouse_step[0] * factor),
            mouse_step[1] - (mouse_step[1] * factor)]

        # Zoom view and move to the mouse positions. Zoom also compensate the
        # delta if center is False.
        mouse_step_out[0] += (
            (current_region.width * factor - current_region.width) / 2)
        mouse_step_out[1] += (
            (current_region.height * factor - current_region.height) / 2)
        return mouse_step_out

    def on_area_button_press_event(self, widget, event):
        if event.button == 1:
            self.mouse_pressed = True
            self.mouse_position = (event.x, event.y)

    def on_area_button_release_event(self, widget, event):
        if event.button == 1:
            self.mouse_pressed = False

    def on_area_motion_notify_event(self, widget, event):
        if self.mouse_pressed:
            self.move(
                self.mouse_position[0] - event.x,
                self.mouse_position[1] - event.y)
        self.mouse_position = (event.x, event.y)

    def on_area_scroll_event(self, widget, event):
        """Callback for the scroll mouse event."""
        if event.direction == gtk.gdk.SCROLL_UP:
            factor = 1.0 / self.zoom_factor
            mouse_step = self._mouse_step_factor(factor)
            self.zoom_in(mouse_step)
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            mouse_step = self._mouse_step_factor(self.zoom_factor)
            self.zoom_out(mouse_step)

    def render(self, context):
        CairoPlayer.render(self, context)
