# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

"""Code to deal with SVG images."""

import json
import math
import re
import rsvg
from glib import GError
from lxml import etree
from numpy import array, dot

from a4lib.region import Region


A4HTMLNS = 'http://launchpad.net/a4/0.1'
# FIXME Get these from the SVG itself, we currently care only about "svg:"
# namespace, the others can be ignored.
a4nsmap = {'a4': A4HTMLNS}
nsmap = {
    'a4': A4HTMLNS,
    'xlink': 'http://www.w3.org/1999/xlink',
    'svg': 'http://www.w3.org/2000/svg'}

available_version = '0.1'


class PresentationError(StandardError):
    """Error raised when an operation with an image fails."""


class NoMetadataError(StandardError):
    """Error raised when an operation with an image fails."""


class Presentation(object):
    """An object that represents a presentation."""

    def __init__(self, file_name, force_loading=False):
        self.file_name = file_name
        try:
            svg_data = open(file_name).read()
            self._svg_handler = rsvg.Handle(data=svg_data)
        except IOError as error:
            raise PresentationError(error.args[1])
        except GError as error:
            raise PresentationError('Unknown file type')

        self.svgtree = etree.parse(file_name)
        self.path = []

        # Get all the text strings contained by our metadata.
        xpath_query = (
            "//a4:presentation")
        presentation_data = self.svgtree.xpath(xpath_query, namespaces=nsmap)
        if (not presentation_data and not force_loading):
            raise NoMetadataError('A4 metadata not found')
        elif (not presentation_data and force_loading):
            presentation_data = self._build_metadata()
        elif (presentation_data[0].get('version') != available_version):
            msg = 'The file you are trying to open has been created with a \
                  version of A4 newer than this one'
            raise PresentationError(msg)
        for child in presentation_data[0]:
            if child.tag == "{%s}shortcuts" % A4HTMLNS:
                ## TODO: handling shortcut
                pass
            if child.tag == "{%s}path" % A4HTMLNS:
                # handling path
                self._build_path(child)

        self.width = self._svg_handler.props.width
        self.height = self._svg_handler.props.height
        self.render_cairo = self._svg_handler.render_cairo

    def _build_metadata(self):
        metadata = etree.SubElement(self.svgtree.getroot(), 'metadata')
        metadata.set('id', 'a4-presentation-information')
        metadata_pres = etree.SubElement(metadata,
                            "{%s}presentation" % A4HTMLNS, nsmap=nsmap)
        metadata_pres.set('version', '0.1')
        path_data = etree.SubElement(metadata_pres,
                            "{%s}path" % A4HTMLNS, nsmap=nsmap)
        return metadata

    def _build_path(self, path_element):
        for child in path_element:
            if child.tag == "{%s}rect" % A4HTMLNS:
                p = PathElement('rect', child.attrib['refers'])
                #props = self.get_item_properties_by_id(child.attrib['refers'])
                #p = Region.from_obj_props(props)
            elif child.tag == "{%s}region" % A4HTMLNS:
                p = Region(float(child.attrib['x']),
                    float(child.attrib['y']),
                    float(child.attrib['width']),
                    float(child.attrib['height']),
                    float(child.attrib['rotate']))
            else:
                continue
            self.path.append(p)

    def _save_path(self, file_name=None):
        xpath_query = (
            "//a4:presentation/a4:path")
        path_data = self.svgtree.xpath(xpath_query, namespaces=nsmap)[0]
        path_data.clear()
        for region in self.path:
            if isinstance(region, Region):
                new_region_tag = etree.Element("{%s}region" % A4HTMLNS)
                new_region_tag.set('x', str(region.xc))
                new_region_tag.set('y', str(region.yc))
                new_region_tag.set('width', str(region.width))
                new_region_tag.set('height', str(region.height))
                new_region_tag.set('rotate', str(region.rotate))
            else:
                new_region_tag = etree.Element("{%s}rect" % A4HTMLNS)
                new_region_tag.set('refers', region.id)
            path_data.append(new_region_tag)
        if file_name:
            out_file = open(file_name, 'w')
        else:
            out_file = open(self.file_name, 'w')
        out_file.write(etree.tostring(self.svgtree, pretty_print=True))

    def _get_region_from_path(self, path_element):
        if isinstance(path_element, Region):
            return path_element
        props = self.get_item_properties_by_id(path_element.id)
        p = Region.from_obj_props(props)
        return p

    def get_path_region(self, index):
        a = self.path[index]
        return self._get_region_from_path(a)

    def get_path_regions(self):
        return [self._get_region_from_path(a) for a in self.path]

    def _handle_transform(self, target_element=None):
        """Handle SVG transform tags and build a list of tranformations."""
        matrix_list = []
        while target_element is not None:
            transform_string = target_element.attrib.get('transform')
            target_element = target_element.getparent()
            if not transform_string:
                # This element doesn't contain a transform tag.
                continue

            # Look for a transformation matrix.
            matches = re.search(
                'matrix\((.*?),(.*?),(.*?),(.*?),(.*?),(.*?)\)',
                transform_string)
            if matches:
                transform_matrix = [
                    float(matches.group(i)) for i in xrange(1, 7)]
                matrix_list.append(transform_matrix)

            # Look for a translation.
            matches = re.search(
                'translate\((.*?),(.*?)\)', transform_string)
            if matches:
                transform_matrix = [
                    1, 0, 0, 1, float(matches.group(1)),
                    float(matches.group(2))]
                matrix_list.append(transform_matrix)

            # Look for a scale factor.
            matches = re.search("scale\(([^,]+)(,([^,]+))?\)",
                transform_string)
            if matches:
                transform_matrix = [
                    float(matches.group(1)), 0, 0,
                    float(matches.group(3) or matches.group(1)), 0, 0]
                matrix_list.append(transform_matrix)

            # Look for a rotation.
            matches = re.search("rotate\((.*?)\)", transform_string)
            if matches:
                angle = math.radians(int(matches.group(1)))
                transform_matrix = [
                    math.cos(angle), math.sin(angle), - math.sin(angle),
                    math.cos(angle), 0, 0]
                matrix_list.append(transform_matrix)

        return matrix_list

    def _reduce_matrix_list(self, mt_list):
        """Take a list of matrixes, as given by SVG code, and returns
            a single numpy.array() representing the whole transformation"""
        if not mt_list:
            return None
        ### reduce the matrix_list to only one.
        ### get the rotation angle (degree?radiants?)
        mt = mt_list[0]
        # head of the list represent the innermost transformation,
        # so we should put it into the end of the dot() reduction.
        head = array(([mt[0], mt[2], mt[4]], [mt[1], mt[3], mt[5]], [0, 0, 1]))
        if len(mt_list) > 1:
            result = dot(self._reduce_matrix_list(mt_list[1:]), head)
        else:
            result = head
        return result

    def _get_rotation_from_matrix(self, matrix):
        """Calculate rotation angle from the transformation matrix."""
        return math.atan2(-matrix[0][1], matrix[0][0])

    def _get_scalefactor_from_matrix(self, matrix=None):
        """ Calculate scale factor from transformation matrix """
        if matrix == None or matrix == []:
            return 1, 1
        scalex = math.sqrt(matrix[0][0] ** 2 + matrix[0][1] ** 2)
        scaley = math.sqrt(matrix[1][0] ** 2 + matrix[1][1] ** 2)
        return scalex, scaley

    def _get_translate_from_matrix(self, matrix=None):
        """ Calculate scale factor from transformation matrix """
        if matrix == None or matrix == []:
            return 0, 0
        scalefactor = math.sqrt(matrix[0][0] ** 2 + matrix[0][1] ** 2)
        angle = -self._get_rotation_from_matrix(matrix)
        x, y = matrix[0][2] / scalefactor, matrix[1][2] / scalefactor
        matrix = [[math.cos(angle), - math.sin(angle), 0],
                  [math.sin(angle), math.cos(angle), 0],
                  [0, 0, 1]]
        x, y, nulla = dot(matrix, [x, y, 1])
        return x, y

    def get_item_properties_by_id(self, item_id):
        """Returns some properties of the SVG element with the given id.

        The properties returned are:
          * the width and the height of the element;
          * its position;
          * the transformations and the rotations.
        """
        try:
            target_element = self.svgtree.xpath(
                "//svg:*[@id='{0}']".format(item_id), namespaces=nsmap)
        except etree.XPathEvalError:
            raise PresentationError('Can find the requested item')

        properties = {}
        if len(target_element) > 0:
            properties['width'] = float(
                target_element[0].attrib.get('width'))
            properties['height'] = float(
                target_element[0].attrib.get('height'))
            properties['x'] = float(target_element[0].attrib.get('x'))
            properties['y'] = float(target_element[0].attrib.get('y'))
            properties['transform'] = self._reduce_matrix_list(
                self._handle_transform(target_element[0]))

            if properties['transform'] != None:
                properties['rotation'] = self._get_rotation_from_matrix(
                    properties['transform'])
                #x,y,null = dot(properties['transform'],
                #    [properties['x'],properties['y'],1])
                #properties['x'] = x
                #properties['y'] = y
            else:
                properties['rotation'] = 0
            properties['scale'] = self._get_scalefactor_from_matrix(
                    properties['transform'])
            properties['translate'] = self._get_translate_from_matrix(
                    properties['transform'])
        return properties

    def get_path(self):
        return self.path


class FeaturedPresentation(Presentation):
    def __init__(self, file_name, force_loading=False):
        Presentation.__init__(self, file_name, force_loading)
        # main_g is the main <g> tag. it will be used mainly to add othe
        # tag painted with A4
        #self.main_g = self.svgtree.xpath('/svg:svg/svg:g', namespaces=nsmap)[0]
        self.main_g = self.svgtree.xpath('/svg:svg', namespaces=nsmap)[0]
        self.render_cairo = self._render_cairo

    def _render_cairo(self, context):
        tempdata = etree.tostring(self.svgtree)
        svg_handler = rsvg.Handle(data=tempdata)
        svg_handler.render_cairo(context)

    def add_object(self, presentation_object):
        self.main_g.append(presentation_object.element)

    def search_id(self, object_id):
        target = self.svgtree.xpath(
                "//*[@id='{0}']".format(piuprops['id']), namespaces=nsmap)[0]
        if target:
            return target
        return False


class PathElement(object):

    def __init__(self, obj_type='rect', obj_id=None):
        self.type = obj_type
        self.id = obj_id
