# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

import random
import math
from lxml import etree

import presentation


class PresentationObject(object):
    def __init__(self, tag_name, x=0, y=0, rotate=0.0):
        self.element = etree.Element(tag_name, nsmap=presentation.nsmap)
        self.element.attrib['id'] = self._generate_new_id()
        self.move(x, y)
        self.rotate(rotate)

    def _generate_new_id(self, length=10):
        seed = "abcdef0123456789"
        return ''.join(random.choice(seed) for i in xrange(length))

    def append_attributes(self, attribs):
        for key in attribs:
            self.element.attrib[key] = str(piuprops[key])

    @property
    def get_position(self):
        return self.element.attrib['x'], self.element.attrib['y']

    @property
    def get_id(self):
        return self.element.attrib['id']

    def move(self, x, y):
        self.element.attrib['x'] = str(x)
        self.element.attrib['y'] = str(y)

    def rotate(self, rotate):
        # need to calculate the transform matrix to rotate the object around his (x,y),
        # otherwise it will rotate around (0,0)
        xc = float(self.element.attrib['x'])
        yc = float(self.element.attrib['y'])
        xc1 = math.cos(-rotate) * xc - math.sin(-rotate) * yc
        yc1 = math.sin(-rotate) * xc + math.cos(-rotate) * yc
        self.element.attrib['transform'] = ("rotate(%s)" % str(math.degrees(rotate))) + \
            (" translate(%s,%s)" % (xc1 - xc, yc1 - yc))

    def set_size(self, width, height):
        self.element.attrib['width'] = str(width)
        self.element.attrib['height'] = str(height)

    ## Methods that could be implemented in subclasses.
    ## Implementing a stub here so subclasses doesn't have to reimplement.
    def start_move(self):
        pass
    def end_move(self):
        pass
    def character_pressed(self, char):
        pass
    def clicked_on_(self, x, y):
        pass


class RectObject(PresentationObject):
    def __init__(self, x=0, y=0, rotate=0.0):
        PresentationObject.__init__(self, 'rect', x, y, rotate)


class TextObject(PresentationObject):
    def __init__(self, x=0, y=0, rotate=0.0, text=""):
        PresentationObject.__init__(self, 'text', x, y, rotate)
        self.element.text = text

    def character_pressed(self, char):
        self.element.text += char

    def get_text(self):
        return self.element.text

    def set_text(self, text):
        self.element.text = text

    text = property(get_text, set_text)


class PolaroidObject(PresentationObject):
    def __init__(self, x=0, y=0, rotate=0.0, text=""):
        PresentationObject.__init__(self, 'g', x, y, rotate)

        self.xratio = 0.95
        self.yratio = 0.80

        self.rect1 = etree.Element('rect')
        self.rect1.attrib['id'] = self._generate_new_id()
        self.rect1.attrib['style'] = "fill:#fbfae1;fill-opacity:1;stroke:#000000;stroke-opacity:1"
        self.rect1.attrib['x'] = str(x)
        self.rect1.attrib['y'] = str(y)

        self.rect2 = etree.Element('rect')
        self.rect2.attrib['id'] = self._generate_new_id()
        self.rect2.attrib['style'] = "fill:#000000;fill-opacity:1;stroke:#000000;stroke-width:1;stroke-miterlimit:4;stroke-opacity:1;stroke-dasharray:none"
        self.rect2.attrib['x'] = str(x)
        self.rect2.attrib['y'] = str(y)

        self.textelement = etree.Element('text')
        self.textelement.attrib['id'] = self._generate_new_id()
        self.textelement.text = ""

        self.element.append(self.rect1)
        self.element.append(self.rect2)
        self.element.append(self.textelement)

    def set_size(self, width, height):
        self.rect1.attrib['width'] = str(width)
        self.rect1.attrib['height'] = str(height)

        self.rect2.attrib['width'] = str(width * self.xratio)
        self.rect2.attrib['height'] = str(height * self.yratio)

        top = (width - width * self.xratio) / 2

        self.rect2.attrib['x'] = str(float(self.rect1.attrib['x']) + top)
        self.rect2.attrib['y'] = str(float(self.rect1.attrib['y']) + top)

        self.textelement.attrib['x'] = str(float(self.rect1.attrib['x']) +
                width / 2)
        self.textelement.attrib['y'] = str(float(self.rect1.attrib['y']) +
                height - (height - (height * self.yratio + top)) / 2)

    def get_size(self):
        return (self.rect1.attrib['width'], self.rect1.attrib['height'])

    def character_pressed(self, char):
        self.textelement.text += char


class ImageObject(PresentationObject):
    def __init__(self, x=0, y=0, rotate=0.0):
        PresentationObject.__init__(self, 'image', x, y, rotate)

    def set_image(self, target):
        href_attr = '{%s}href' % "http://www.w3.org/1999/xlink"
        self.element.attrib[href_attr] = "file://" + target
