# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

"""Code to handle regions of interest."""

import math
import time


class Region(object):
    """A region of interest in the presentation.

    A region of interest is subset of the image that we may want to show.
    Regions of interest are always rotated rectangles.
    """

    def __init__(self, xc, yc, width, height, rotate):
        self.xc = xc
        self.yc = yc
        self.width = width
        self.height = height
        self.rotate = rotate

    def __str__(self):
        return "Region (%d,%d) - %dx%d - rad: %f" % (self.xc, self.yc,
               self.width, self.height, self.rotate)

    @classmethod
    def whole_image(cls, image):
        """Initialize a region that covers the whole image."""
        w = float(image.width)
        h = float(image.height)
        return cls(w / 2, h / 2, w, h, 0.0)

    @classmethod
    def from_obj_props(cls, obj_props):
        """Initialize from a given object's properties."""
        o = obj_props
        xc = (o['x'] + o['translate'][0] + o['width'] / 2) * o['scale'][0]
        yc = (o['y'] + o['translate'][1] + o['height'] / 2) * o['scale'][1]
        w = o['width'] * o['scale'][0]
        h = o['height'] * o['scale'][1]
        r = o['rotation']
        xc1 = math.cos(r) * xc - math.sin(r) * yc
        yc1 = math.sin(r) * xc + math.cos(r) * yc
        return cls(xc1, yc1, w, h, r)

    def scale_of(self, scalex, scaley):
        """Returns a `Region` that's a scaled version of self."""
        return Region(self.xc, self.yc, self.width * scalex,
            self.height * scaley, self.rotate)

    def move(self, stepx, stepy):
        stepx1 = math.cos(self.rotate) * stepx - math.sin(self.rotate) * stepy
        stepy1 = math.sin(self.rotate) * stepx + math.cos(self.rotate) * stepy
        return Region(
            self.xc + stepx1, self.yc + stepy1, self.width, self.height,
            self.rotate)

    def get_window_scale_vector(self, window_size):
        """ Returns the two ratio between a given size and the current
            instance."""
        win_width, win_height = window_size
        scale_x = 1.0 * win_width / self.width
        scale_y = 1.0 * win_height / self.height
        return (scale_x, scale_y)

    def get_window_scale_factor(self, window_size):
        """ Returns a single scale factor, the one that should be applied when
            handling a region in the cairo context."""
        scale_x, scale_y = self.get_window_scale_vector(window_size)
        if scale_x < scale_y:
            return scale_x
        else:
            return scale_y

    def get_window_fit_vector(self, window_size):
        """ Returns the blank border that should be applied to a Region in
            order to fit in a windows of different aspect ratio."""
        scale_x, scale_y = self.get_window_scale_vector(window_size)
        if scale_x < scale_y:
            return (0, (scale_y - scale_x) * self.height)
        else:
            return ((scale_x - scale_y) * self.width, 0)

    def get_point_from_window_coordinates(self, window_size, window_point):
        scale = self.get_window_scale_factor(window_size)
        translate = self.get_window_fit_vector(window_size)

        x = window_point[0]
        y = window_point[1]

        x -= translate[0] / 2.0
        y -= translate[1] / 2.0

        x = x / scale
        y = y / scale

        x -= self.width / 2
        y -= self.height / 2
        x1 = math.cos(self.rotate) * x - math.sin(self.rotate) * y
        y1 = math.sin(self.rotate) * x + math.cos(self.rotate) * y
        x1 += self.xc
        y1 += self.yc

        return (x1, y1)

    def render(self, context, svg_image, window_size):
        """Render the region of interest to the given Cairo context."""
        scale = self.get_window_scale_factor(window_size)
        translate = self.get_window_fit_vector(window_size)
        translate = translate[0] / 2.0, translate[1] / 2.0
        # First transform of context, that handle the aspect of the window:
        #   we don't want to deform our region and keep it centered in the
        #   context.
        context.translate(translate[0], translate[1])
        context.scale(scale, scale)
        # Really handle the transformation of the region
        context.translate(self.width / 2, self.height / 2)
        context.rotate(-self.rotate)
        context.translate(-self.xc, -self.yc)
        svg_image.render_cairo(context)


def linear_interpolation(fraction):
    """Constant speed interpolation."""
    return fraction


def quadratic_interpolation(fraction):
    """Create an elastic movement animation."""
    now0 = fraction * 2
    if now0 < 1:
        # Manage the ease-in acceleration.
        out = now0 * now0 / 2
    else:
        # Manage the ease-out deceleration.
        now0 -= 1
        out = -(now0 * (now0 - 2) - 1) / 2
    return out


class Animation(object):

    def __init__(
        self, startregion, endregion, deltatime=1.0,
        interpolation_type=quadratic_interpolation):
        self.startregion = startregion
        self.endregion = endregion
        self.time = deltatime
        self.interpolator = interpolation_type
        self.status = 'stopped'

    def _interpolate(self, now):
        fraction = (now - self.starttime) / self.time
        fraction = self.interpolator(fraction)
        start, end = self.startregion, self.endregion
        xc = start.xc + fraction * (end.xc - start.xc)
        yc = start.yc + fraction * (end.yc - start.yc)
        width = start.width + fraction * (end.width - start.width)
        height = start.height + fraction * (end.height - start.height)
        deltarot = (end.rotate - start.rotate) % (2 * math.pi)
        if deltarot > math.pi:
            deltarot = deltarot - 2 * math.pi
        rotate = start.rotate + fraction * deltarot
        return Region(xc, yc, width, height, rotate)

    def get_next_region(self):
        now = time.time()
        if self.status == 'stopped':
            self.starttime = now
            self.status = 'started'
        if now > (self.starttime + self.time):
            self.status = 'ended'
            return self.endregion
        return self._interpolate(now)
