# Copyright 2010 A4 Developers.  This software is licensed under the
# GNU General Public License version 3 (see the file COPYING).

"""Helpers containing code common to almost every the test."""

import os
import ldtp

# An image to be used for tests.
TEST_IMAGE_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'images', 'roi.svg'))

# The default timeout for `waittillguiexist` and `waittillguinotexist`.
_DEFAULT_TIMEOUT = 1


class Window(object):
    """An object representing a window or a message dialog."""

    def __init__(self, window_id):
        self.window_id = window_id
        assert ldtp.waittillguiexist(window_id, guiTimeOut=_DEFAULT_TIMEOUT)

    def __getattr__(self, attr_name):
        """Return a wrapper around the LDTP method with the given name.

        The returned wrapper is a function that automatically send the window
        ID to the `ldtp` module.
        """
        ldtp_method = getattr(ldtp, attr_name)

        def wrapper(*args, **kwargs):
            retval = ldtp_method(self.window_id, *args, **kwargs)
            if isinstance(retval, int):
                assert retval
            else:
                return retval

        return wrapper

    def __del__(self):
        """Wait for the closure of the window."""
        # `window_id` may be unset if __init__ did not complete successfully.
        if hasattr(self, 'window_id'):
            ldtp.closewindow(self.window_id)
            ldtp.waittillguinotexist(
                self.window_id, guiTimeOut=_DEFAULT_TIMEOUT)

    def __repr__(self):
        return '<{0} {1!r}>'.format(self.__class__.__name__, self.window_id)


class Application(Window):
    """The main window of A4."""

    def __init__(self, *args):
        assert ldtp.launchapp('./a4', args)
        super(Application, self).__init__('frmA4')

    @property
    def window_list(self):
        """Return a list of IDs of the open windows, including this one."""
        return ldtp.getwindowlist()

    def wait_for(self, window_id):
        """Wait for a window to be opened and return it's object."""
        return Window(window_id)
